### COLORS ###

lilypond_version = "2.19.82"

normal_colors = set(
    [
        "black",
        "white",
        "red",
        "green",
        "blue",
        "cyan",
        "magenta",
        "yellow",
        "gray",
        "grey",
        "darkred",
        "darkgreen",
        "darkblue",
        "darkcyan",
        "darkmagenta",
        "darkyellow",
    ]
)

colors = set(normal_colors)

colors.update(
    [
        "snow",
        "GhostWhite",
        "WhiteSmoke",
        "gainsboro",
        "FloralWhite",
        "OldLace",
        "linen",
        "AntiqueWhite",
        "PapayaWhip",
        "BlanchedAlmond",
        "bisque",
        "PeachPuff",
        "NavajoWhite",
        "moccasin",
        "cornsilk",
        "ivory",
        "LemonChiffon",
        "seashell",
        "honeydew",
        "MintCream",
        "azure",
        "AliceBlue",
        "lavender",
        "LavenderBlush",
        "MistyRose",
        "white",
        "black",
        "DarkSlateGrey",
        "DimGrey",
        "SlateGrey",
        "LightSlateGrey",
        "grey",
        "LightGrey",
        "MidnightBlue",
        "navy",
        "NavyBlue",
        "CornflowerBlue",
        "DarkSlateBlue",
        "SlateBlue",
        "MediumSlateBlue",
        "LightSlateBlue",
        "MediumBlue",
        "RoyalBlue",
        "blue",
        "DodgerBlue",
        "DeepSkyBlue",
        "SkyBlue",
        "LightSkyBlue",
        "SteelBlue",
        "LightSteelBlue",
        "LightBlue",
        "PowderBlue",
        "PaleTurquoise",
        "DarkTurquoise",
        "MediumTurquoise",
        "turquoise",
        "cyan",
        "LightCyan",
        "CadetBlue",
        "MediumAquamarine",
        "aquamarine",
        "DarkGreen",
        "DarkOliveGreen",
        "DarkSeaGreen",
        "SeaGreen",
        "MediumSeaGreen",
        "LightSeaGreen",
        "PaleGreen",
        "SpringGreen",
        "LawnGreen",
        "green",
        "chartreuse",
        "MediumSpringGreen",
        "GreenYellow",
        "LimeGreen",
        "YellowGreen",
        "ForestGreen",
        "OliveDrab",
        "DarkKhaki",
        "khaki",
        "PaleGoldenrod",
        "LightGoldenrodYellow",
        "LightYellow",
        "yellow",
        "gold",
        "LightGoldenrod",
        "goldenrod",
        "DarkGoldenrod",
        "RosyBrown",
        "IndianRed",
        "SaddleBrown",
        "sienna",
        "peru",
        "burlywood",
        "beige",
        "wheat",
        "SandyBrown",
        "tan",
        "chocolate",
        "firebrick",
        "brown",
        "DarkSalmon",
        "salmon",
        "LightSalmon",
        "orange",
        "DarkOrange",
        "coral",
        "LightCoral",
        "tomato",
        "OrangeRed",
        "red",
        "HotPink",
        "DeepPink",
        "pink",
        "LightPink",
        "PaleVioletRed",
        "maroon",
        "MediumVioletRed",
        "VioletRed",
        "magenta",
        "violet",
        "plum",
        "orchid",
        "MediumOrchid",
        "DarkOrchid",
        "DarkViolet",
        "BlueViolet",
        "purple",
        "MediumPurple",
        "thistle",
        "DarkGrey",
        "DarkBlue",
        "DarkCyan",
        "DarkMagenta",
        "DarkRed",
        "LightGreen",
    ]
)

for string in (
    "snow",
    "seashell",
    "AntiqueWhite",
    "bisque",
    "PeachPuff",
    "NavajoWhite",
    "LemonChiffon",
    "cornsilk",
    "ivory",
    "honeydew",
    "LavenderBlush",
    "MistyRose",
    "azure",
    "SlateBlue",
    "RoyalBlue",
    "blue",
    "DodgerBlue",
    "SteelBlue",
    "DeepSkyBlue",
    "SkyBlue",
    "LightSkyBlue",
    "LightSteelBlue",
    "LightBlue",
    "LightCyan",
    "PaleTurquoise",
    "CadetBlue",
    "turquoise",
    "cyan",
    "aquamarine",
    "DarkSeaGreen",
    "SeaGreen",
    "PaleGreen",
    "SpringGreen",
    "green",
    "chartreuse",
    "OliveDrab",
    "DarkOliveGreen",
    "khaki",
    "LightGoldenrod",
    "LightYellow",
    "yellow",
    "gold",
    "goldenrod",
    "DarkGoldenrod",
    "RosyBrown",
    "IndianRed",
    "sienna",
    "burlywood",
    "wheat",
    "tan",
    "chocolate",
    "firebrick",
    "brown",
    "salmon",
    "LightSalmon",
    "orange",
    "DarkOrange",
    "coral",
    "tomato",
    "OrangeRed",
    "red",
    "DeepPink",
    "HotPink",
    "pink",
    "LightPink",
    "PaleVioletRed",
    "maroon",
    "VioletRed",
    "magenta",
    "orchid",
    "plum",
    "MediumOrchid",
    "DarkOrchid",
    "purple",
    "MediumPurple",
    "thistle",
):
    for i in range(4):
        n = i + 1
        color = string + str(n)
        colors.add(color)

for i in range(101):
    color = "gray" + str(i)
    colors.add(color)
    color = "grey" + str(i)
    colors.add(color)

### DRUMS ###

lilypond_version = "2.16.1"

drums = {
    "acousticbassdrum": "acousticbassdrum",
    "acousticsnare": "acousticsnare",
    "agh": "hiagogo",
    "agl": "loagogo",
    "bassdrum": "bassdrum",
    "bd": "bassdrum",
    "bda": "acousticbassdrum",
    "boh": "hibongo",
    "bohm": "mutehibongo",
    "boho": "openhibongo",
    "bol": "lobongo",
    "bolm": "mutelobongo",
    "bolo": "openlobongo",
    "cab": "cabasa",
    "cabasa": "cabasa",
    "cb": "cowbell",
    "cgh": "hiconga",
    "cghm": "mutehiconga",
    "cgho": "openhiconga",
    "cgl": "loconga",
    "cglm": "muteloconga",
    "cglo": "openloconga",
    "chinesecymbal": "chinesecymbal",
    "cl": "claves",
    "claves": "claves",
    "closedhihat": "closedhihat",
    "cowbell": "cowbell",
    "crashcymbal": "crashcymbal",
    "crashcymbala": "crashcymbala",
    "crashcymbalb": "crashcymbalb",
    "cuim": "mutecuica",
    "cuio": "opencuica",
    "cymc": "crashcymbal",
    "cymca": "crashcymbala",
    "cymcb": "crashcymbalb",
    "cymch": "chinesecymbal",
    "cymr": "ridecymbal",
    "cymra": "ridecymbala",
    "cymrb": "ridecymbalb",
    "cyms": "splashcymbal",
    "da": "onedown",
    "db": "twodown",
    "dc": "threedown",
    "dd": "fourdown",
    "de": "fivedown",
    "electricsnare": "electricsnare",
    "fivedown": "fivedown",
    "fiveup": "fiveup",
    "fourdown": "fourdown",
    "fourup": "fourup",
    "gui": "guiro",
    "guil": "longguiro",
    "guiro": "guiro",
    "guis": "shortguiro",
    "halfopenhihat": "halfopenhihat",
    "handclap": "handclap",
    "hc": "handclap",
    "hh": "hihat",
    "hhc": "closedhihat",
    "hhho": "halfopenhihat",
    "hho": "openhihat",
    "hhp": "pedalhihat",
    "hiagogo": "hiagogo",
    "hibongo": "hibongo",
    "hiconga": "hiconga",
    "highfloortom": "highfloortom",
    "hightom": "hightom",
    "hihat": "hihat",
    "himidtom": "himidtom",
    "hisidestick": "hisidestick",
    "hitimbale": "hitimbale",
    "hiwoodblock": "hiwoodblock",
    "loagogo": "loagogo",
    "lobongo": "lobongo",
    "loconga": "loconga",
    "longguiro": "longguiro",
    "longwhistle": "longwhistle",
    "losidestick": "losidestick",
    "lotimbale": "lotimbale",
    "lowfloortom": "lowfloortom",
    "lowmidtom": "lowmidtom",
    "lowoodblock": "lowoodblock",
    "lowtom": "lowtom",
    "mar": "maracas",
    "maracas": "maracas",
    "mutecuica": "mutecuica",
    "mutehibongo": "mutehibongo",
    "mutehiconga": "mutehiconga",
    "mutelobongo": "mutelobongo",
    "muteloconga": "muteloconga",
    "mutetriangle": "mutetriangle",
    "onedown": "onedown",
    "oneup": "oneup",
    "opencuica": "opencuica",
    "openhibongo": "openhibongo",
    "openhiconga": "openhiconga",
    "openhihat": "openhihat",
    "openlobongo": "openlobongo",
    "openloconga": "openloconga",
    "opentriangle": "opentriangle",
    "pedalhihat": "pedalhihat",
    "rb": "ridebell",
    "ridebell": "ridebell",
    "ridecymbal": "ridecymbal",
    "ridecymbala": "ridecymbala",
    "ridecymbalb": "ridecymbalb",
    "shortguiro": "shortguiro",
    "shortwhistle": "shortwhistle",
    "sidestick": "sidestick",
    "sn": "snare",
    "sna": "acousticsnare",
    "snare": "snare",
    "sne": "electricsnare",
    "splashcymbal": "splashcymbal",
    "ss": "sidestick",
    "ssh": "hisidestick",
    "ssl": "losidestick",
    "tamb": "tambourine",
    "tambourine": "tambourine",
    "threedown": "threedown",
    "threeup": "threeup",
    "timh": "hitimbale",
    "timl": "lotimbale",
    "tomfh": "highfloortom",
    "tomfl": "lowfloortom",
    "tomh": "hightom",
    "toml": "lowtom",
    "tommh": "himidtom",
    "tomml": "lowmidtom",
    "tri": "triangle",
    "triangle": "triangle",
    "trim": "mutetriangle",
    "trio": "opentriangle",
    "tt": "tamtam",
    "twodown": "twodown",
    "twoup": "twoup",
    "ua": "oneup",
    "ub": "twoup",
    "uc": "threeup",
    "ud": "fourup",
    "ue": "fiveup",
    "vibraslap": "vibraslap",
    "vibs": "vibraslap",
    "wbh": "hiwoodblock",
    "wbl": "lowoodblock",
    "whl": "longwhistle",
    "whs": "shortwhistle",
}
